/** *****************************************************************************************
 * Unit in charge:
 * @file	vehspdctrl.c
 * @author	Brieske
 * $Date:: 2017-09-12 08:12:11 #$
 *
 * ----------------------------------------------------------
 * SVN information of last commit:
 * $Rev:: 2022                  $
 * $Author:: Brieske            $
 *
 * ----------------------------------------------------------
 *
 * @brief TODO
 *
 ******************************************************************************************/

/************* Includes *******************************************************************/
#include "../vehspdctrl.h"
#include "app/stm/stm.h"
#include "dev/rcrec/rcrec.h"
#include "dev/servo/servo.h"
#include "app/leddebug.h"

/************* Global variables *******************************************************************/
/* input signals */
float *pVehSpdOdom_vehspdctrl;
uint8_t *pState_vehspdctrl;
carcontrol_t *pCarControl_vehspdctrl;
//float *pVehSpdTgtTraj_vehspdctrl;
uint16_t *pRcChannel1_vehspdctrl;

/* output signals */
//float *pMotTrqTgtSrv_vehspdctrl;

/* parameters */

/************* Private typedefs ***********************************************************/

/************* Macros and constants *******************************************************/
#define TDELTA 0.02

#define PID_K_P 0.2
#define PID_K_I 0.2
#define PID_K_D 0
//#define PID_K_P 0.1
//#define PID_K_I 0.16
//#define PID_K_D 0.0
#define PID_GAIN 1

#define PT1_T 0.08
#define PT1_T_OPT (1. / (PT1_T / TDELTA) + 1)

#define MOT_MAXVAL 10.
#define U_DEADBAND_POSITIVE 0.0583
#define U_DEADBAND_NEGATIVE -0.075

/************* Private static variables ***************************************************/
static pidCtrl_t pid;
static pt1_t pt1;

static float vehSpdOdom;
static uint8_t state;
static float motTrqTgtSrv;

/************* Private function prototypes ************************************************/

float spdCtrl(float w);

/************* Public functions ***********************************************************/
int8_t vehspdctrl_init(void)
{
	pCarControl_vehspdctrl->vehSpdTgtExt = 0.0f;

	pt1.K = 1;
	pt1.T_opt = PT1_T_OPT;
	pt1.y_old = 0;

	pid.kp = PID_K_P;
	pid.ki = PID_K_I;
	pid.kd = PID_K_D;

//	pid.ki = 0.F;
//	pid.kd = 0.F;

	pid.gain = PID_GAIN;
	pid.ta = TDELTA;
	pid.min = -MOT_MAXVAL;
	pid.max = MOT_MAXVAL;

	pid_initialize(&pid);

	return 0;
}

void vehspdctrl_step(void)
{
	irq_disable();
	state = *pState_vehspdctrl;
	irq_enable();

	switch (state) {
	case SYSTEM_STATE_INITIALIZING:
		// motor is disabled, nothing to do here
		return;
	case SYSTEM_STATE_IDLE:
		motTrqTgtSrv = 0;
	case SYSTEM_STATE_EMERGENCY:
		motTrqTgtSrv = spdCtrl(0.F);
		break;
	case SYSTEM_STATE_RUNNING_RC:
		irq_disable();
		float rcvalue = (float) *pRcChannel1_vehspdctrl;
		irq_enable();
//		if (rcvalue < 1400) {
//			motTrqTgtSrv = 0;
//		}
//		else {
			motTrqTgtSrv = (rcvalue - 1500.0F) / (float) (MAX_PERIOD_CHANNEL - MIN_PERIOD_CHANNEL);
			motTrqTgtSrv *= 0.8;
//			motTrqTgtSrv *= 3;
//			motTrqTgtSrv = spdCtrl(motTrqTgtSrv);
//		}
		break;
	case SYSTEM_STATE_RUNNING_EXT:
		irq_disable();
		float spdTgt = pCarControl_vehspdctrl->vehSpdTgtExt;
		irq_enable();

		if (spdTgt != 0.0f)
			debugLedOn();
		else
			debugLedOff();
			
		motTrqTgtSrv = spdCtrl(spdTgt);
		break;
//	case SYSTEM_STATE_RUNNING_TRAJ:
//		TODO system state not implemented yet
//		irq_disable();
//		float spdTgt = *pVehSpdTgtTraj_vehspdctrl;
//		irq_enable();
//		motTrqTgtSrv = spdCtrl(spdTgt);
//		break
	default:
		// cannot go here
		break;
	}

	servo_updateMotorTrq(motTrqTgtSrv);
}

/************* Private functions **********************************************************/

float spdCtrl(float w)
{
	float u = 0;	// manipulated value
	float u_f = 0;	// feed forward value
	float y = 0;	// controlled value

	irq_disable();
	y = -*pVehSpdOdom_vehspdctrl;
	irq_enable();

	y = pt1_calculate(&pt1, y);
	pid_controller(&pid, y, w, &u, -MOT_MAXVAL, MOT_MAXVAL);

	if (u > 0) {
		u_f = U_DEADBAND_POSITIVE;
	}
	else if (u < 0) {
		u_f = U_DEADBAND_NEGATIVE;
	}

	u += u_f;

	return u;
}
